/+
	Copyright (c) 2005 Eric Anderton
	        
	Permission is hereby granted, free of charge, to any person
	obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without
	restriction, including without limitation the rights to use,
	copy, modify, merge, publish, distribute, sublicense, and/or
	sell copies of the Software, and to permit persons to whom the
	Software is furnished to do so, subject to the following
	conditions:

	The above copyright notice and this permission notice shall be
	included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
	EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
	NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
	HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
	FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
	OTHER DEALINGS IN THE SOFTWARE.
+/
// DDL bless utility
module utils.bless;

private import ddl.all;
private import ddl.FileBuffer;
private import ddl.Attributes;
private import ddl.DDLException;
private import ddl.DDLReader;

private import ddl.ddl.DDLLoader;
private import ddl.ddl.DDLBinary;

private import mango.io.File;
private import mango.io.Stdout;
private import mango.io.FilePath;
private import mango.io.FileConduit;

private import mango.text.LineIterator;

private import utils.bless_bn;
private import utils.ArgParser;

enum PathPart{
	Ext = 1,
	Name = 2,
	Path = 4,
	Root = 8,
	Suffix = 16
}

char[] filePathPart(FilePath path,PathPart parts){
	char[] result;
	
	if(parts | PathPart.Root) result ~= path.getRoot;
	if(parts | PathPart.Path) result ~= path.getPath;
	if(parts | PathPart.Name) result ~= path.getName;
	if(parts | PathPart.Suffix) result ~= path.getSuffix;
	if(parts | PathPart.Ext) result ~= path.getExtension;
	
	return result;	
}


char[] helpText = 
"Bless - DDL wrapper utility - V1.2 Build %d
Copyright (C) 2005,2006 Eric Anderton
Documentation: http://www.dsource.org/projects/ddl

Creates a .ddl file that wraps the provided object file.
The new file name will adopt the name of the provided
object file with a 'ddl' extension by default.

Usage: 
  bless <object file> { -switch }

  -v       verbose output
  -x       (curse) extract object file from .ddl
  -n       no-create (for test purposes)
  -f<file> output to file 'file' (overrides default name)
  -a<attr> adds an attribute (name=value) to the ddl
  -p<file> uses the property file 'file' for attributes
  
";

int main(char[][] args){
	DefaultRegistry registry = new DefaultRegistry();
	
	if(args.length == 1){
		Stdout.println(helpText,auto_build_number);
		Stdout.println("Supported Object Types:");
		foreach(char[] ext; registry.getSupportedTypes){
			Stdout.print("%s ",ext);	
		}
		Stdout.println("");
		return 0;
	}
		
	// get the filename

	// pick through the options (if any)
	bit verbose = false;
	bit extract = false;
	bit noCreate = false;
	char[] outputFileName = "";
	char[] filename = "";
	Attributes attributes;
	
	// configure the arg parser
	ArgParser parser = new ArgParser(delegate uint(char[] value,uint ordinal){
		if(ordinal > 0) throw new DDLException("Invalid argument %s",value);
		filename = value;
		return value.length;
	});
	
	parser.bind("-", "v",delegate void(){
		verbose=true;
	});
	
	parser.bind("-", "n",delegate void(){
		noCreate=true;
	});	
	
	parser.bind("-", "x",delegate void(){
		extract=true;
	});	
		
	parser.bind("-", "f",delegate uint(char[] value){
		outputFileName=value;
		return value.length;
	});
	
	char[] trim(char[] token){
		uint start;
		uint end;
		for(start=0; start<token.length && token[start] <= 32; start++){} // leading whitespace
		for(end=token.length-1; end>=0 && token[end] <= 32; end--){} // trailing whitespace
		return token[start..end+1];
	}
	
	void parseAttribute(char[] line){
		if(line.length == 0 || line[0] == '#') return;
		uint split;
		for(split=0; split<line.length && line[split] != '='; split++){} // token
		if(split < line.length-1){
			attributes[trim(line[0..split])] = trim(line[split+1..$]);
		}
	}
	
	parser.bind("-", "a",delegate uint(char[] value){
		parseAttribute(value);
		return value.length;
	});
	
	parser.bind("-", "p",delegate uint(char[] value){
		FileConduit file = new FileConduit(value);
		//TODO: determine if file exists, or just let exception fall through?
		//if(file.exists){
			LineIterator iter = new LineIterator(file);
			while(iter.next){
				parseAttribute(iter.get);
			}
			return value.length;
		//}
		//else{
		//	throw new Exception("Cannot find properties file: '" ~ value ~ "'");
		//}
	});
	
	// parse the arguments
	parser.parse(args[1..$]);
		
	if(filename == ""){
		throw new DDLException("No filename specified.");
	}
	
	// get the file
	FileBuffer file = new FileBuffer(filename);
	char[] originalOutputFile = filePathPart(file.getPath,PathPart.Root | PathPart.Path | PathPart.Name | PathPart.Suffix);
	
	if(outputFileName == ""){
		outputFileName = originalOutputFile;
	}
	
	if(noCreate){
		Stdout.println("loaded attributes:");
		foreach(char[] name,char[] value; attributes){
			Stdout.println("%s = %s",name,value);
		}
	}
	
	// validate a few things
	/*
	if(!file.exists){
		Stdout.println("File '%s' does not exist.",file.filename);
		return 1;
	}
	*/
	
	// load it up!
	//NOTE: rather than just be satisfied with what loader likes this file
	// we load the entire file via its loader to ensure that it will work later.
	DynamicLibrary lib = registry.load(file.getPath.toString());
		
	if(!lib){
		Stdout.println("Cannot load '%s'",file.getPath.toString());
		return 1;
	}
		
	// output
	if(verbose){
		Stdout.println("filename: '%s'",file.getPath.toString());
		Stdout.println("type: '%s'",lib.getType);
	}	
		
	if(extract){
		if(lib.getType != DDLLoader.typeName){
			Stdout.println("cannot extract from a non-ddl file ('%s').",file.getPath.toString());
			return 1;
		}
	}	
	else if(lib.getType == DDLLoader.typeName){
		Stdout.println("'%s' already appears to be a ddl file.",file.getPath.toString());
		return 1;
	}
	
	if(extract){
		// coerce complete load of the ddl binary file
		DDLBinary binary = new DDLBinary();
		binary.load(new FileBuffer(file.getPath));
		
		// save to output file
		if(outputFileName == originalOutputFile){
			DynamicLibraryLoader loader = registry.getLoader(binary.binaryType);
			if("std.filename" in binary.attributes){
				outputFileName = binary.attributes["std.filename"];
			}
			else{
				outputFileName = originalOutputFile ~ ".bin";
			}
		}
				
		if(verbose) Stdout.println("out file: '%s'",outputFileName);
		
		if(!noCreate){
			File destFile = new File(outputFileName);
			destFile.write(binary.getBinaryData);
			if(verbose) Stdout.println("Created '%s'",outputFileName);
		}
	}
	else{
		if(verbose) Stdout.println("out file: '%s'",outputFileName);

		// create the file and save it
		DDLBinary binary = new DDLBinary();
		DDLReader reader = new DDLReader(file);
		void[] data;

		// save the easy fields	
		binary.binaryType = lib.getType;
		binary.attributes = attributes;		
		
		reader.getAll(data);
		binary.binaryData = cast(ubyte[])data;
								
		if(!("std.filename" in attributes)){
			attributes["std.filename"] = file.getPath.getName ~ "." ~ file.getPath.getExtension;
		}

		// get the defined namespaces and imported modules
		foreach(DynamicModule mod; lib.getModules()){
			DemangleType demangleType;
			char[] result;

			foreach(sym; mod.getSymbols){
				demangleType = getDemangleType(sym.name);
				if(demangleType == DemangleType.ModuleInfo){
					if(sym.type == SymbolType.Strong || SymbolType.Weak){					
						binary.definedNamespaces ~= sym.name;
						if(verbose) Stdout.println("module %s",demangleSymbol(sym.name));
					}
					else{
						binary.importedModules ~= sym.name;
						if(verbose) Stdout.println("module %s",demangleSymbol(sym.name));
					}
				}
			}
		}
				
		if(!noCreate){
			FileBuffer outputFile = new FileBuffer(outputFileName,FileStyle.ReadWriteCreate);
			binary.save(outputFile);
						
			outputFile.flush();
			outputFile.getConduit().close();
			
			if(verbose) Stdout.println("Created '%s'",outputFileName);
		}
	}
	
	return 0;
}