/+
	Copyright (c) 2005-2006 Eric Anderton
	        
	Permission is hereby granted, free of charge, to any person
	obtaining a copy of this software and associated documentation
	files (the "Software"), to deal in the Software without
	restriction, including without limitation the rights to use,
	copy, modify, merge, publish, distribute, sublicense, and/or
	sell copies of the Software, and to permit persons to whom the
	Software is furnished to do so, subject to the following
	conditions:

	The above copyright notice and this permission notice shall be
	included in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
	EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
	OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
	NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
	HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
	FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
	OTHER DEALINGS IN THE SOFTWARE.
+/
/// DDL info utility
module utils.ddlinfo;

private import utils.ddlinfo_bn;
private import utils.ArgParser;

private import ddl.all;
private import ddl.DDLException;
private import ddl.FileBuffer;

private import mango.io.Stdout;

char[] helpText = 
"DDL Info - DDL information utility - V1.2 Build %d
Copyright (C) 2005-2006 Eric Anderton
Documentation: http://www.dsource.org/projects/ddl

Displays information about a ddl or object file.

Usage: 
  ddlinfo <object file> { -switch }
  
  -r       raw output instead of friendly names/symbols.
  -a       attributes only
  -s       symbols only
  -x	   x-ray - show *everything*
";

int main(char[][] args){
	DefaultRegistry regsitry = new DefaultRegistry();
	
	if(args.length == 1){
		Stdout.println(helpText,auto_build_number);
		Stdout.println("Supported Object Types:");
		foreach(char[] ext; regsitry.getSupportedTypes){
			Stdout.print("%s ",ext);	
		}
		Stdout.println("");
		return 0;
	}
		
	// pick through the options (if any)
	bool rawOutput = false;
	bool displayAttribs = false;
	bool displaySymbols = false;
	bool xRay = false;
	char[] filename = "";
	
	// configure the arg parser
	ArgParser parser = new ArgParser(delegate uint(char[] value,uint ordinal){
		if(ordinal > 0) throw new DDLException("Invalid argument %s",value);
		filename = value;
		return value.length;
	});
			
	parser.bind("-", "r",delegate void(){
		rawOutput=true;
	});
	
	parser.bind("-", "a",delegate void(){
		displayAttribs=true;
	});
	
	parser.bind("-", "s",delegate void(){
		displaySymbols=true;
	});
	
	parser.bind("-", "x",delegate void(){
		xRay=true;
	});
	
	// parse the arguments
	parser.parse(args[1..$]);

	if(!displaySymbols && !displayAttribs){
		displaySymbols = true;
		displayAttribs = true;
	}
	
	if(filename == ""){
		throw new DDLException("No filename specified.");
	}
	
	// get the file
	FileBuffer file = new FileBuffer(filename);	

	//TODO: validate a few things
/*	if(!file.exists){
		Stdout.println("file '%s' does not exist.",file.filename);
		return 1;
	}
*/		
	// (attempt to) load the binary file
	DynamicLibrary lib = regsitry.load(file);
	
	
	if(!lib){
		Stdout.println("file '%s' is not supported.",file.getPath.toString());
		return 1;
	}

	Stdout.println("filename: '%s'",file.getPath.toString());
	Stdout.println("type: '%s'",lib.getType);
	
	if(xRay){
		Stdout.println("%s",lib.toString());
	}
	else{
		if(displayAttribs){
			if(lib.getAttributes.length > 0){
				Stdout.println("attributes: ");
				foreach(char[] name,char[] value; lib.getAttributes){
					Stdout.println("%s - %s",name,value);
				}
			}
		}
		if(displaySymbols){
			Stdout.println("\nModules (%d):",lib.getModules.length);
			foreach(DynamicModule mod; lib.getModules()){
				Stdout.println("\n%s",mod.getName());
				Stdout.println("\nSymbols (%d):",mod.getSymbols.length);
				
				foreach(symbol; mod.getSymbols){
					if(rawOutput){
						Stdout.println("%s %s",symbol.getTypeName,symbol.name);
					}
					else{
						Stdout.println("%s %s",symbol.getTypeName,demangleSymbol(symbol.name));
					}				
				}
			}
		}
	}

	return 0;
}